"""Module to load and init namespace at package level"""
from .chembl import ChEMBL
from .chemidplus import ChemIDplus
from .drugbank import DrugBank
from .drugsatfda import DrugsAtFDA
from .guidetopharmacology import GuideToPHARMACOLOGY
from .hemonc import HemOnc
from .ncit import NCIt
from .rxnorm import RxNorm
from .wikidata import Wikidata
from .merge import Merge
